@@ -1332,6 +1332,22 @@ do_video_rotate_action (CsdMediaKeysManager *manager,
         return FALSE;
 }
 
+static void
+do_video_rotate_lock_action (CsdMediaKeysManager *manager,
+                             gint64               timestamp)
+{
+        GSettings *settings;
+        gboolean locked;
+
+        settings = g_settings_new ("org.cinnamon.settings-daemon.peripherals.touchscreen");
+        locked = !g_settings_get_boolean (settings, "orientation-lock");
+        g_settings_set_boolean (settings, "orientation-lock", locked);
+        g_object_unref (settings);
+
+        show_osd (manager, locked ? "rotation-locked-symbolic"
+                                  : "rotation-allowed-symbolic", NULL, -1);
+}
+
 static void
 do_toggle_accessibility_key (const char *key)
 {
@@ -1732,6 +1748,9 @@ do_action (CsdMediaKeysManager *manager,
         case C_DESKTOP_MEDIA_KEY_ROTATE_VIDEO:
                 do_video_rotate_action (manager, timestamp);
                 break;
+        case C_DESKTOP_MEDIA_KEY_ROTATE_VIDEO_LOCK:
+                do_video_rotate_lock_action (manager, timestamp);
+                break;
         case C_DESKTOP_MEDIA_KEY_SCREENREADER:
                 do_screenreader_action (manager);
                 break;
