@@ -64,12 +64,15 @@
 
 /* Touchpad settings */
 #define KEY_TOUCHPAD_DISABLE_W_TYPING    "disable-while-typing"
-#define KEY_PAD_HORIZ_SCROLL             "horiz-scroll-enabled"
-#define KEY_SCROLL_METHOD                "scroll-method"
 #define KEY_TAP_TO_CLICK                 "tap-to-click"
 #define KEY_TWO_FINGER_CLICK             "two-finger-click"
 #define KEY_THREE_FINGER_CLICK           "three-finger-click"
 
+#define KEY_VERT_EDGE_SCROLL             "vertical-edge-scrolling"
+#define KEY_HORIZ_EDGE_SCROLL            "horizontal-edge-scrolling"
+#define KEY_VERT_TWO_FINGER_SCROLL       "vertical-two-finger-scrolling"
+#define KEY_HORIZ_TWO_FINGER_SCROLL      "horizontal-two-finger-scrolling"
+
 #define KEY_TOUCHPAD_ENABLED             "touchpad-enabled"
 #define KEY_NATURAL_SCROLL_ENABLED       "natural-scroll"
 
@@ -308,6 +311,16 @@ bail:
         return FALSE;
 }
 
+static Atom
+property_from_name (const char *property_name)
+{
+        Atom prop;
+
+        prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), property_name, True);
+
+        return prop;
+}
+
 static gboolean
 touchpad_has_single_button (XDevice *device)
 {
@@ -318,7 +331,7 @@ touchpad_has_single_button (XDevice *device)
         gboolean is_single_button = FALSE;
         int rc;
 
-        prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Capabilities", False);
+        prop = property_from_name ("Synaptics Capabilities");
         if (!prop)
                 return FALSE;
 
@@ -337,11 +350,103 @@ touchpad_has_single_button (XDevice *device)
         return is_single_button;
 }
 
+static gboolean
+property_exists_on_device (GdkDevice *device, const char * property_name)
+{
+        int rc;
+        Atom act_type, property;
+        int act_format;
+        unsigned long nitems, bytes_after;
+        unsigned char *data;
+        XDevice *xdevice;
+
+        property = property_from_name (property_name);
+        if (!property)
+                return FALSE;
+
+        xdevice = open_gdk_device (device);
+        if (xdevice == NULL)
+                return FALSE;
+
+        gdk_error_trap_push ();
+        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
+                                 property, 0, 1, False,
+                                 XA_INTEGER, &act_type, &act_format, &nitems,
+                                 &bytes_after, &data);
+        if (rc == Success)
+            XFree(data);
+
+        gdk_error_trap_pop_ignored ();
+
+        return rc == Success;
+}
+
 static void
-set_left_handed (CsdMouseManager *manager,
-                 GdkDevice       *device,
-                 gboolean mouse_left_handed,
-                 gboolean touchpad_left_handed)
+property_set_bool (GdkDevice *device,
+                   XDevice *xdevice,
+                   const char * property_name,
+                   int property_index,
+                   gboolean enable)
+{
+        int rc;
+        Atom act_type, property;
+        int act_format;
+        unsigned long nitems, bytes_after;
+        unsigned char *data;
+
+        property = property_from_name (property_name);
+        if (!property) {
+                return;
+        }
+
+        int value = 0;
+        if (enable) {
+                value = 1;
+        }
+
+        g_debug ("Setting %s on %s to %d", property_name, gdk_device_get_name (device), value);
+
+        gdk_error_trap_push ();
+        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
+                                 property, 0, 1, False,
+                                 XA_INTEGER, &act_type, &act_format, &nitems,
+                                 &bytes_after, &data);
+        if (rc == Success && act_type == XA_INTEGER && act_format == 8 && nitems > property_index) {
+                data[property_index] = value;
+                XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
+                                       property, XA_INTEGER, 8,
+                                       PropModeReplace, data, nitems);
+        }
+
+        if (rc == Success) {
+                XFree (data);
+        }
+
+        if (gdk_error_trap_pop ()) {
+                g_warning ("Error while setting %s on \"%s\"", property_name, gdk_device_get_name (device));
+        }
+}
+
+static void
+touchpad_set_bool (GdkDevice *device, const char * property_name, int property_index, gboolean enable)
+{
+        XDevice *xdevice;
+
+        xdevice = open_gdk_device (device);
+        if (xdevice == NULL)
+                return;
+
+        if (device_is_touchpad (xdevice))
+                property_set_bool (device, xdevice, property_name, property_index, enable);
+
+        xdevice_close (xdevice);
+}
+
+static void
+set_left_handed_legacy_driver (CsdMouseManager *manager,
+                               GdkDevice       *device,
+                               gboolean mouse_left_handed,
+                               gboolean touchpad_left_handed)
 {
         XDevice *xdevice;
         guchar *buttons;
@@ -356,7 +461,7 @@ set_left_handed (CsdMouseManager *manager,
         if (xdevice == NULL)
                 return;
 
-	g_debug ("setting handedness on %s", gdk_device_get_name (device));
+        g_debug ("setting handedness on %s", gdk_device_get_name (device));
 
         buttons = g_new (guchar, buttons_capacity);
 
@@ -393,7 +498,7 @@ set_left_handed (CsdMouseManager *manager,
 
         configure_button_layout (buttons, n_buttons, left_handed);
 
-	gdk_error_trap_push ();
+        gdk_error_trap_push ();
         XSetDeviceButtonMapping (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice, buttons, n_buttons);
         gdk_error_trap_pop_ignored ();
 
@@ -403,8 +508,39 @@ out:
 }
 
 static void
-set_motion (CsdMouseManager *manager,
-            GdkDevice       *device)
+set_left_handed_libinput (GdkDevice       *device,
+                          gboolean mouse_left_handed,
+                          gboolean touchpad_left_handed)
+{
+        XDevice *xdevice;
+        gboolean want_lefthanded;
+
+        xdevice = open_gdk_device (device);
+        if (xdevice == NULL)
+                return;
+
+        want_lefthanded = device_is_touchpad (xdevice) ? touchpad_left_handed : mouse_left_handed;
+
+        property_set_bool (device, xdevice, "libinput Left Handed Enabled", 0, want_lefthanded);
+
+        xdevice_close (xdevice);
+}
+
+static void
+set_left_handed (CsdMouseManager *manager,
+                 GdkDevice       *device,
+                 gboolean mouse_left_handed,
+                 gboolean touchpad_left_handed)
+{
+        if (property_exists_on_device (device, "libinput Left Handed Enabled"))
+                set_left_handed_libinput (device, mouse_left_handed, touchpad_left_handed);
+        else
+                set_left_handed_legacy_driver (manager, device, mouse_left_handed, touchpad_left_handed);
+}
+
+static void
+set_motion_legacy_driver (CsdMouseManager *manager,
+                          GdkDevice       *device)
 {
         XDevice *xdevice;
         XPtrFeedbackControl feedback;
@@ -420,7 +556,7 @@ set_motion (CsdMouseManager *manager,
         if (xdevice == NULL)
                 return;
 
-	g_debug ("setting motion on %s", gdk_device_get_name (device));
+        g_debug ("setting motion on %s", gdk_device_get_name (device));
 
         if (device_is_touchpad (xdevice))
                 settings = manager->priv->touchpad_settings;
@@ -494,9 +630,97 @@ set_motion (CsdMouseManager *manager,
 }
 
 static void
-set_middle_button (CsdMouseManager *manager,
-                   GdkDevice       *device,
-                   gboolean         middle_button)
+set_motion_libinput (CsdMouseManager *manager,
+                     GdkDevice       *device)
+{
+        int rc;
+        XDevice *xdevice;
+        GSettings *settings;
+        Atom act_type, property, float_type;
+        int act_format;
+        unsigned long nitems, bytes_after;
+        union {
+                unsigned char *c;
+                long *l;
+        } data;
+        gfloat accel;
+        gfloat motion_acceleration;
+
+        xdevice = open_gdk_device (device);
+        if (xdevice == NULL)
+                return;
+
+        g_debug ("setting motion on %s", gdk_device_get_name (device));
+
+        if (device_is_touchpad (xdevice))
+                settings = manager->priv->touchpad_settings;
+        else
+                settings = manager->priv->mouse_settings;
+
+
+        /* Calculate acceleration */
+        motion_acceleration = g_settings_get_double (settings, KEY_MOTION_ACCELERATION);
+
+        /* panel gives us a range of 1.0-10.0, map to libinput's -1, 1
+
+           oldrange = (oldmax - oldmin)
+           newrange = (newmax - newmin)
+
+           mapped = (value - oldmin) * newrange / oldrange + oldmin
+         */
+
+        if (motion_acceleration == -1.0) /* unset */
+                accel = 0.0;
+        else
+                accel = (motion_acceleration - 1.0) * 2.0 / 9.0 - 1;
+
+
+        float_type = property_from_name ("FLOAT");
+        if (!float_type)
+                return;
+
+        property = property_from_name ("libinput Accel Speed");
+        if (!property) {
+                return;
+        }
+
+        gdk_error_trap_push ();
+        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
+                                 property, 0, 1, False,
+                                 float_type, &act_type, &act_format, &nitems,
+                                 &bytes_after, &data.c);
+        if (rc == Success && act_type == float_type && act_format == 32 && nitems >= 1) {
+                *(float*)data.l = accel;
+                XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
+                                       property, float_type, 32,
+                                       PropModeReplace, data.c, nitems);
+        }
+
+        if (rc == Success) {
+                XFree (data.c);
+        }
+
+        if (gdk_error_trap_pop ()) {
+                g_warning ("Error while setting accel speed on \"%s\"", gdk_device_get_name (device));
+        }
+
+        xdevice_close (xdevice);
+}
+
+static void
+set_motion (CsdMouseManager *manager,
+            GdkDevice       *device)
+{
+        if (property_exists_on_device (device, "libinput Accel Speed"))
+                set_motion_libinput (manager, device);
+        else
+                set_motion_legacy_driver (manager, device);
+}
+
+static void
+set_middle_button_evdev (CsdMouseManager *manager,
+                         GdkDevice       *device,
+                         gboolean         middle_button)
 {
         Atom prop;
         XDevice *xdevice;
@@ -506,8 +730,7 @@ set_middle_button (CsdMouseManager *manager,
         unsigned char *data;
         int rc;
 
-        prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
-                            "Evdev Middle Button Emulation", True);
+        prop = property_from_name ("Evdev Middle Button Emulation");
 
         if (!prop) /* no evdev devices */
                 return;
@@ -516,7 +739,7 @@ set_middle_button (CsdMouseManager *manager,
         if (xdevice == NULL)
                 return;
 
-	g_debug ("setting middle button on %s", gdk_device_get_name (device));
+        g_debug ("setting middle button on %s", gdk_device_get_name (device));
 
         gdk_error_trap_push ();
 
@@ -540,6 +763,41 @@ set_middle_button (CsdMouseManager *manager,
         xdevice_close (xdevice);
 }
 
+static void
+set_middle_button_libinput (CsdMouseManager *manager,
+                            GdkDevice       *device,
+                            gboolean         middle_button)
+{
+        XDevice *xdevice;
+
+        g_debug ("setting middle button on %s", gdk_device_get_name (device));
+
+        xdevice = open_gdk_device (device);
+        if (xdevice == NULL)
+                return;
+
+        /* we didn't set it for synaptics, so bail out for touchpads */
+        if (device_is_touchpad (xdevice))
+                return;
+
+        property_set_bool (device, xdevice, "libinput Middle Emulation Enabled",
+                           0, middle_button);
+
+        xdevice_close (xdevice);
+}
+
+static void
+set_middle_button (CsdMouseManager *manager,
+                   GdkDevice       *device,
+                   gboolean         middle_button)
+{
+        if (property_from_name ("Evdev Middle Button Emulation"))
+                set_middle_button_evdev (manager, device, middle_button);
+
+        if (property_from_name ("libinput Middle Emulation Enabled"))
+                set_middle_button_libinput (manager, device, middle_button);
+}
+
 /* Ensure that syndaemon dies together with us, to avoid running several of
  * them */
 static void
@@ -573,7 +831,7 @@ syndaemon_died (GPid pid, gint status, gpointer user_data)
 }
 
 static int
-set_disable_w_typing (CsdMouseManager *manager, gboolean state)
+set_disable_w_typing_synaptics (CsdMouseManager *manager, gboolean state)
 {
         if (state && touchpad_is_present ()) {
                 GError *error = NULL;
@@ -624,10 +882,45 @@ set_disable_w_typing (CsdMouseManager *manager, gboolean state)
         return 0;
 }
 
+static int
+set_disable_w_typing_libinput (CsdMouseManager *manager, gboolean state)
+{
+        GList *devices, *l;
+
+        /* This is only called once for synaptics but for libinput we need
+         * to loop through the list of devices
+         */
+        devices = gdk_device_manager_list_devices (manager->priv->device_manager, GDK_DEVICE_TYPE_SLAVE);
+
+        for (l = devices; l != NULL; l = l->next) {
+                GdkDevice *device = l->data;
+
+                if (device_is_ignored (manager, device))
+                        continue;
+
+                touchpad_set_bool (device, "libinput Disable While Typing Enabled", 0, state);
+        }
+        g_list_free (devices);
+
+        return 0;
+}
+
+static int
+set_disable_w_typing (CsdMouseManager *manager, gboolean state)
+{
+        if (property_from_name ("Synaptics Off"))
+                set_disable_w_typing_synaptics (manager, state);
+
+        if (property_from_name ("libinput Disable While Typing Enabled"))
+                set_disable_w_typing_libinput (manager, state);
+
+        return 0;
+}
+
 static void
-set_tap_to_click (GdkDevice *device,
-                  gboolean   state,
-                  gboolean   left_handed)
+set_tap_to_click_synaptics (GdkDevice *device,
+                            gboolean   state,
+                            gboolean   left_handed)
 {
         int format, rc;
         unsigned long nitems, bytes_after;
@@ -635,7 +928,7 @@ set_tap_to_click (GdkDevice *device,
         unsigned char* data;
         Atom prop, type;
 
-        prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Tap Action", False);
+        prop = property_from_name ("Synaptics Tap Action");
         if (!prop)
                 return;
 
@@ -648,7 +941,7 @@ set_tap_to_click (GdkDevice *device,
                 return;
         }
 
-	g_debug ("setting tap to click on %s", gdk_device_get_name (device));
+        g_debug ("setting tap to click on %s", gdk_device_get_name (device));
 
         gdk_error_trap_push ();
         rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice, prop, 0, 2,
@@ -656,9 +949,9 @@ set_tap_to_click (GdkDevice *device,
                                  &bytes_after, &data);
 
         if (rc == Success && type == XA_INTEGER && format == 8 && nitems >= 7) {
-               /* Set MR mapping for corner tapping on the right side*/
-               data[0] = (state) ? 2 : 0;
-               data[1] = (state) ? 3 : 0;
+                /* Set MR mapping for corner tapping on the right side*/
+                data[0] = (state) ? 2 : 0;
+                data[1] = (state) ? 3 : 0;
 
                 /* Set RLM mapping for 1/2/3 fingers*/
                 data[4] = (state) ? ((left_handed) ? 3 : 1) : 0;
@@ -678,9 +971,29 @@ set_tap_to_click (GdkDevice *device,
 }
 
 static void
-set_click_actions (GdkDevice *device,
-                   gint   enable_two_finger_click,
-                   gint   enable_three_finger_click)
+set_tap_to_click_libinput (GdkDevice *device, gboolean state)
+{
+        g_debug ("setting tap to click on %s", gdk_device_get_name (device));
+
+        touchpad_set_bool (device, "libinput Tapping Enabled", 0, state);
+}
+
+static void
+set_tap_to_click (GdkDevice *device,
+                  gboolean   state,
+                  gboolean   left_handed)
+{
+        if (property_from_name ("Synaptics Tap Action"))
+                set_tap_to_click_synaptics (device, state, left_handed);
+
+        if (property_from_name ("libinput Tapping Enabled"))
+                set_tap_to_click_libinput (device, state);
+}
+
+static void
+set_click_actions_synaptics (GdkDevice *device,
+                             gint   enable_two_finger_click,
+                             gint   enable_three_finger_click)
 {
         int format, rc;
         unsigned long nitems, bytes_after;
@@ -688,7 +1001,7 @@ set_click_actions (GdkDevice *device,
         unsigned char* data;
         Atom prop, type;
 
-        prop = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Click Action", False);
+        prop = property_from_name ("Synaptics Click Action");
         if (!prop)
                 return;
 
@@ -701,7 +1014,7 @@ set_click_actions (GdkDevice *device,
                 return;
         }
 
-    g_debug ("setting click action to click on %s", gdk_device_get_name (device));
+        g_debug ("setting click action to click on %s", gdk_device_get_name (device));
 
         gdk_error_trap_push ();
         rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice, prop, 0, 2,
@@ -712,7 +1025,8 @@ set_click_actions (GdkDevice *device,
             data[0] = 1;
             data[1] = enable_two_finger_click;
             data[2] = enable_three_finger_click;
-            XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice, prop, XA_INTEGER, 8, PropModeReplace, data, nitems);
+            XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
+                                   prop, XA_INTEGER, 8, PropModeReplace, data, nitems);
         }
 
         if (rc == Success)
@@ -725,20 +1039,19 @@ set_click_actions (GdkDevice *device,
 }
 
 static void
-set_horiz_scroll (GdkDevice *device,
-                  gboolean   state)
+set_click_actions_libinput (GdkDevice *device,
+                            gint   enable_two_finger_click,
+                            gint   enable_three_finger_click)
 {
-        int rc;
-        XDevice *xdevice;
-        Atom act_type, prop_edge, prop_twofinger;
-        int act_format;
+        int format, rc;
         unsigned long nitems, bytes_after;
-        unsigned char *data;
+        XDevice *xdevice;
+        unsigned char* data;
+        Atom prop, type;
+        gboolean want_softwarebuttons, want_clickfinger;
 
-        prop_edge = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Edge Scrolling", False);
-        prop_twofinger = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Two-Finger Scrolling", False);
-
-        if (!prop_edge || !prop_twofinger)
+        prop = property_from_name ("libinput Click Method Enabled");
+        if (!prop)
                 return;
 
         xdevice = open_gdk_device (device);
@@ -750,60 +1063,65 @@ set_horiz_scroll (GdkDevice *device,
                 return;
         }
 
-	g_debug ("setting horiz scroll on %s", gdk_device_get_name (device));
+        g_debug ("setting click action to click on %s", gdk_device_get_name (device));
+
+        want_clickfinger = enable_two_finger_click || enable_three_finger_click;
+        want_softwarebuttons = !want_clickfinger;
 
         gdk_error_trap_push ();
-        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
-                                 prop_edge, 0, 1, False,
-                                 XA_INTEGER, &act_type, &act_format, &nitems,
+        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice, prop, 0, 2,
+                                 False, XA_INTEGER, &type, &format, &nitems,
                                  &bytes_after, &data);
-        if (rc == Success && act_type == XA_INTEGER &&
-            act_format == 8 && nitems >= 2) {
-                data[1] = (state && data[0]);
-                XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
-                                       prop_edge, XA_INTEGER, 8,
-                                       PropModeReplace, data, nitems);
+
+        if (rc == Success && type == XA_INTEGER && format == 8 && nitems >= 2) {
+                data[0] = want_softwarebuttons;
+                data[1] = want_clickfinger;
+                XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice, prop, XA_INTEGER, 8, PropModeReplace, data, nitems);
         }
 
-        XFree (data);
-
-        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
-                                 prop_twofinger, 0, 1, False,
-                                 XA_INTEGER, &act_type, &act_format, &nitems,
-                                 &bytes_after, &data);
-        if (rc == Success && act_type == XA_INTEGER &&
-            act_format == 8 && nitems >= 2) {
-                data[1] = (state && data[0]);
-                XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
-                                       prop_twofinger, XA_INTEGER, 8,
-                                       PropModeReplace, data, nitems);
-        }
-
-        if (gdk_error_trap_pop ())
-                g_warning ("Error in setting horiz scroll on \"%s\"", gdk_device_get_name (device));
-
         if (rc == Success)
                 XFree (data);
 
-        xdevice_close (xdevice);
+        if (gdk_error_trap_pop ())
+                g_warning ("Error in setting click actions on \"%s\"", gdk_device_get_name (device));
 
+        xdevice_close (xdevice);
 }
 
 static void
-set_edge_scroll (GdkDevice               *device,
-                 CsdTouchpadScrollMethod  method)
+set_click_actions (GdkDevice *device,
+                   gint   enable_two_finger_click,
+                   gint   enable_three_finger_click)
 {
-        int rc;
-        XDevice *xdevice;
-        Atom act_type, prop_edge, prop_twofinger;
-        int act_format;
+        if (property_from_name ("Synaptics Click Action"))
+                set_click_actions_synaptics (device, enable_two_finger_click, enable_three_finger_click);
+
+        if (property_from_name ("libinput Click Method Enabled"))
+                set_click_actions_libinput (device, enable_two_finger_click, enable_three_finger_click);
+}
+
+static void
+set_scrolling_synaptics (GdkDevice *device, GSettings *settings)
+{
+        touchpad_set_bool (device, "Synaptics Edge Scrolling", 0, g_settings_get_boolean (settings, KEY_VERT_EDGE_SCROLL));
+        touchpad_set_bool (device, "Synaptics Edge Scrolling", 1, g_settings_get_boolean (settings, KEY_HORIZ_EDGE_SCROLL));
+        touchpad_set_bool (device, "Synaptics Two-Finger Scrolling", 0, g_settings_get_boolean (settings, KEY_VERT_TWO_FINGER_SCROLL));
+        touchpad_set_bool (device, "Synaptics Two-Finger Scrolling", 1, g_settings_get_boolean (settings, KEY_HORIZ_TWO_FINGER_SCROLL));
+}
+
+static void
+set_scrolling_libinput (GdkDevice *device, GSettings *settings)
+{
+        int format, rc;
         unsigned long nitems, bytes_after;
-        unsigned char *data;
+        XDevice *xdevice;
+        unsigned char* data;
+        Atom prop, type;
+        gboolean want_edge, want_2fg;
+        gboolean want_horiz;
 
-        prop_edge = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Edge Scrolling", False);
-        prop_twofinger = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), "Synaptics Two-Finger Scrolling", False);
-
-        if (!prop_edge || !prop_twofinger)
+        prop = property_from_name ("libinput Scroll Method Enabled");
+        if (!prop)
                 return;
 
         xdevice = open_gdk_device (device);
@@ -815,42 +1133,58 @@ set_edge_scroll (GdkDevice               *device,
                 return;
         }
 
-	g_debug ("setting edge scroll on %s", gdk_device_get_name (device));
+        want_2fg = g_settings_get_boolean (settings, KEY_VERT_TWO_FINGER_SCROLL);
+        want_edge = g_settings_get_boolean (settings, KEY_VERT_EDGE_SCROLL);
+
+        /* libinput only allows for one scroll method at a time. If both are
+         * set, pick 2fg scrolling */
+        if (want_2fg)
+                want_edge = FALSE;
+
+        g_debug ("setting scroll method on %s", gdk_device_get_name (device));
 
         gdk_error_trap_push ();
-        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
-                                 prop_edge, 0, 1, False,
-                                 XA_INTEGER, &act_type, &act_format, &nitems,
+        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice, prop, 0, 2,
+                                 False, XA_INTEGER, &type, &format, &nitems,
                                  &bytes_after, &data);
-        if (rc == Success && act_type == XA_INTEGER &&
-            act_format == 8 && nitems >= 2) {
-                data[0] = (method == CSD_TOUCHPAD_SCROLL_METHOD_EDGE_SCROLLING) ? 1 : 0;
+
+        if (rc == Success && type == XA_INTEGER && format == 8 && nitems >= 3) {
+                data[0] = want_2fg;
+                data[1] = want_edge;
                 XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
-                                       prop_edge, XA_INTEGER, 8,
-                                       PropModeReplace, data, nitems);
+                                       prop, XA_INTEGER, 8, PropModeReplace, data, nitems);
         }
 
-        XFree (data);
-
-        rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
-                                 prop_twofinger, 0, 1, False,
-                                 XA_INTEGER, &act_type, &act_format, &nitems,
-                                 &bytes_after, &data);
-        if (rc == Success && act_type == XA_INTEGER &&
-            act_format == 8 && nitems >= 2) {
-                data[0] = (method == CSD_TOUCHPAD_SCROLL_METHOD_TWO_FINGER_SCROLLING) ? 1 : 0;
-                XChangeDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
-                                       prop_twofinger, XA_INTEGER, 8,
-                                       PropModeReplace, data, nitems);
-        }
-
-        if (gdk_error_trap_pop ())
-                g_warning ("Error in setting edge scroll on \"%s\"", gdk_device_get_name (device));
-
         if (rc == Success)
                 XFree (data);
 
+        if (gdk_error_trap_pop ())
+                g_warning ("Error in setting scroll method on \"%s\"", gdk_device_get_name (device));
+
         xdevice_close (xdevice);
+
+        /* horizontal scrolling is handled by xf86-input-libinput and
+         * there's only one bool. Pick the one matching the scroll method we
+         * picked above.
+         */
+        if (want_2fg)
+                want_horiz = g_settings_get_boolean (settings, KEY_HORIZ_TWO_FINGER_SCROLL);
+        else if (want_edge)
+                want_horiz = g_settings_get_boolean (settings, KEY_HORIZ_EDGE_SCROLL);
+        else
+                return;
+
+        touchpad_set_bool (device, "libinput Horizontal Scroll Enabled", 0, want_horiz);
+}
+
+static void
+set_scrolling (GdkDevice *device, GSettings *settings)
+{
+        if (property_from_name ("Synaptics Edge Scrolling"))
+                set_scrolling_synaptics (device, settings);
+
+        if (property_from_name ("libinput Scroll Method Enabled"))
+                set_scrolling_libinput (device, settings);
 }
 
 static void
@@ -1015,17 +1349,16 @@ set_mouse_settings (CsdMouseManager *manager,
 
         set_tap_to_click (device, g_settings_get_boolean (manager->priv->touchpad_settings, KEY_TAP_TO_CLICK), touchpad_left_handed);
         set_click_actions( device, g_settings_get_int (manager->priv->touchpad_settings, KEY_TWO_FINGER_CLICK), g_settings_get_int (manager->priv->touchpad_settings, KEY_THREE_FINGER_CLICK));
-        set_edge_scroll (device, g_settings_get_enum (manager->priv->touchpad_settings, KEY_SCROLL_METHOD));
-        set_horiz_scroll (device, g_settings_get_boolean (manager->priv->touchpad_settings, KEY_PAD_HORIZ_SCROLL));
+        set_scrolling (device, manager->priv->touchpad_settings);
         set_natural_scroll (manager, device, g_settings_get_boolean (manager->priv->touchpad_settings, KEY_NATURAL_SCROLL_ENABLED));
         if (g_settings_get_boolean (manager->priv->touchpad_settings, KEY_TOUCHPAD_ENABLED) == FALSE)
                 set_touchpad_disabled (device);
 }
 
 static void
-set_natural_scroll (CsdMouseManager *manager,
-                    GdkDevice       *device,
-                    gboolean         natural_scroll)
+set_natural_scroll_synaptics (CsdMouseManager *manager,
+                              GdkDevice       *device,
+                              gboolean         natural_scroll)
 {
         XDevice *xdevice;
         Atom scrolling_distance, act_type;
@@ -1047,8 +1380,7 @@ set_natural_scroll (CsdMouseManager *manager,
                  natural_scroll ? "natural (reverse) scroll" : "normal scroll",
                  gdk_device_get_name (device));
 
-        scrolling_distance = XInternAtom (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()),
-                                          "Synaptics Scrolling Distance", False);
+        scrolling_distance = property_from_name ("Synaptics Scrolling Distance");
 
         gdk_error_trap_push ();
         rc = XGetDeviceProperty (GDK_DISPLAY_XDISPLAY (gdk_display_get_default ()), xdevice,
@@ -1083,6 +1415,30 @@ set_natural_scroll (CsdMouseManager *manager,
         xdevice_close (xdevice);
 }
 
+static void
+set_natural_scroll_libinput (CsdMouseManager *manager,
+                             GdkDevice       *device,
+                             gboolean         natural_scroll)
+{
+        g_debug ("Trying to set %s for \"%s\"",
+                 natural_scroll ? "natural (reverse) scroll" : "normal scroll",
+                 gdk_device_get_name (device));
+
+        touchpad_set_bool (device, "libinput Natural Scrolling Enabled", 0, natural_scroll);
+}
+
+static void
+set_natural_scroll (CsdMouseManager *manager,
+                    GdkDevice       *device,
+                    gboolean         natural_scroll)
+{
+        if (property_from_name ("Synaptics Scrolling Distance"))
+                set_natural_scroll_synaptics (manager, device, natural_scroll);
+
+        if (property_from_name ("libinput Natural Scrolling Enabled"))
+                set_natural_scroll_libinput (manager, device, natural_scroll);
+}
+
 static void
 mouse_callback (GSettings       *settings,
                 const gchar     *key,
@@ -1144,16 +1500,15 @@ touchpad_callback (GSettings       *settings,
                         continue;
 
                 if (g_str_equal (key, KEY_TAP_TO_CLICK)) {
+                        gboolean mouse_left_handed;
+                        mouse_left_handed = g_settings_get_boolean (manager->priv->mouse_settings, KEY_LEFT_HANDED);
                         set_tap_to_click (device, g_settings_get_boolean (settings, key),
-                                          g_settings_get_boolean (manager->priv->touchpad_settings, KEY_LEFT_HANDED));
+                                          get_touchpad_handedness (manager, mouse_left_handed));
                 }
                 else if (g_str_equal (key, KEY_TWO_FINGER_CLICK) || g_str_equal (key, KEY_THREE_FINGER_CLICK)) {
                         set_click_actions( device, g_settings_get_int (manager->priv->touchpad_settings, KEY_TWO_FINGER_CLICK), g_settings_get_int (manager->priv->touchpad_settings, KEY_THREE_FINGER_CLICK));
-                } else if (g_str_equal (key, KEY_SCROLL_METHOD)) {
-                        set_edge_scroll (device, g_settings_get_enum (settings, key));
-                        set_horiz_scroll (device, g_settings_get_boolean (settings, KEY_PAD_HORIZ_SCROLL));
-                } else if (g_str_equal (key, KEY_PAD_HORIZ_SCROLL)) {
-                        set_horiz_scroll (device, g_settings_get_boolean (settings, key));
+                } else if (g_str_equal (key, KEY_VERT_EDGE_SCROLL) || g_str_equal (key, KEY_HORIZ_EDGE_SCROLL) || g_str_equal (key, KEY_VERT_TWO_FINGER_SCROLL) || g_str_equal (key, KEY_HORIZ_TWO_FINGER_SCROLL)) {
+                        set_scrolling (device, settings);
                 } else if (g_str_equal (key, KEY_TOUCHPAD_ENABLED)) {
                         if (g_settings_get_boolean (settings, key) == FALSE)
                                 set_touchpad_disabled (device);
