@@ -27,8 +27,6 @@
 #include <glib.h>
 #include <gtk/gtk.h>
 #include <gdk/gdk.h>
-#include <gdk/gdkx.h>
-#include <gudev/gudev.h>
 
 #define GNOME_DESKTOP_USE_UNSTABLE_API
 #include <libcinnamon-desktop/gnome-rr.h>
@@ -49,20 +47,18 @@ typedef enum {
 
 struct CsdOrientationManagerPrivate
 {
-        guint start_idle_id;
 
         /* Accelerometer */
-        char *sysfs_path;
-        OrientationUp prev_orientation;
+        guint          watch_id;
+        GDBusProxy    *iio_proxy;
+        gboolean       has_accel;
+        OrientationUp  prev_orientation;
 
         /* DBus */
-        GDBusNodeInfo   *introspection_data;
-        GDBusConnection *connection;
         GDBusProxy      *xrandr_proxy;
         GCancellable    *cancellable;
 
         /* Notifications */
-        GUdevClient *client;
         GSettings *settings;
         gboolean orientation_lock;
 };
@@ -70,15 +66,9 @@ struct CsdOrientationManagerPrivate
 #define CONF_SCHEMA "org.cinnamon.settings-daemon.peripherals.touchscreen"
 #define ORIENTATION_LOCK_KEY "orientation-lock"
 
+#define CSD_DBUS_NAME "org.cinnamon.SettingsDaemon"
 #define CSD_DBUS_PATH "/org/cinnamon/SettingsDaemon"
-#define CSD_ORIENTATION_DBUS_PATH CSD_DBUS_PATH "/Orientation"
-
-static const gchar introspection_xml[] =
-"<node>"
-"  <interface name='org.cinnamon.SettingsDaemon.Orientation'>"
-"    <annotation name='org.freedesktop.DBus.GLib.CSymbol' value='csd_orientation_manager'/>"
-"  </interface>"
-"</node>";
+#define CSD_DBUS_BASE_INTERFACE "org.cinnamon.SettingsDaemon"
 
 static void     csd_orientation_manager_class_init  (CsdOrientationManagerClass *klass);
 static void     csd_orientation_manager_init        (CsdOrientationManager      *orientation_manager);
@@ -96,33 +86,11 @@ static char *mpu6050_accel_x = NULL;
 static char *mpu6050_accel_y = NULL;
 static gboolean mpu_timer(CsdOrientationManager *manager);
 
-static GObject *
-csd_orientation_manager_constructor (GType                     type,
-                               guint                      n_construct_properties,
-                               GObjectConstructParam     *construct_properties)
-{
-        CsdOrientationManager      *orientation_manager;
-
-        orientation_manager = CSD_ORIENTATION_MANAGER (G_OBJECT_CLASS (csd_orientation_manager_parent_class)->constructor (type,
-                                                                                                         n_construct_properties,
-                                                                                                         construct_properties));
-
-        return G_OBJECT (orientation_manager);
-}
-
-static void
-csd_orientation_manager_dispose (GObject *object)
-{
-        G_OBJECT_CLASS (csd_orientation_manager_parent_class)->dispose (object);
-}
-
 static void
 csd_orientation_manager_class_init (CsdOrientationManagerClass *klass)
 {
         GObjectClass   *object_class = G_OBJECT_CLASS (klass);
 
-        object_class->constructor = csd_orientation_manager_constructor;
-        object_class->dispose = csd_orientation_manager_dispose;
         object_class->finalize = csd_orientation_manager_finalize;
 
         g_type_class_add_private (klass, sizeof (CsdOrientationManagerPrivate));
@@ -187,20 +155,21 @@ orientation_to_string (OrientationUp o)
 }
 
 static OrientationUp
-get_orientation_from_device (GUdevDevice *dev)
+get_orientation_from_device (CsdOrientationManager *manager)
 {
-        const char *value;
+        GVariant *v;
+        OrientationUp o;
 
-        value = g_udev_device_get_property (dev, "ID_INPUT_ACCELEROMETER_ORIENTATION");
-        if (value == NULL) {
-                g_debug ("Couldn't find orientation for accelerometer %s",
-                         g_udev_device_get_sysfs_path (dev));
+        v = g_dbus_proxy_get_cached_property (manager->priv->iio_proxy, "AccelerometerOrientation");
+        if (v == NULL) {
+                g_debug ("Couldn't find orientation for accelerometer");
                 return ORIENTATION_UNDEFINED;
         }
-        g_debug ("Found orientation '%s' for accelerometer %s",
-                 value, g_udev_device_get_sysfs_path (dev));
+        g_debug ("Found orientation '%s' for accelerometer", g_variant_get_string (v, NULL));
 
-        return orientation_from_string (value);
+        o = orientation_from_string (g_variant_get_string (v, NULL));
+        g_variant_unref (v);
+        return o;
 }
 
 static void
@@ -213,10 +182,9 @@ on_xrandr_action_call_finished (GObject               *source_object,
 
         variant = g_dbus_proxy_call_finish (G_DBUS_PROXY (source_object), res, &error);
 
-        g_object_unref (manager->priv->cancellable);
-        manager->priv->cancellable = NULL;
+        g_clear_object (&manager->priv->cancellable);
 
-        if (error != NULL) {
+        if (variant == NULL) {
                 g_warning ("Unable to call 'RotateTo': %s", error->message);
                 g_error_free (error);
         } else {
@@ -232,7 +200,7 @@ do_xrandr_action (CsdOrientationManager *manager,
         GTimeVal tv;
         gint64 timestamp;
 
-        if (priv->connection == NULL || priv->xrandr_proxy == NULL) {
+        if (priv->xrandr_proxy == NULL) {
                 g_warning ("No existing D-Bus connection trying to handle XRANDR keys");
                 return;
         }
@@ -276,39 +244,6 @@ do_rotation (CsdOrientationManager *manager)
         do_xrandr_action (manager, rotation);
 }
 
-static void
-client_uevent_cb (GUdevClient           *client,
-                  gchar                 *action,
-                  GUdevDevice           *device,
-                  CsdOrientationManager *manager)
-{
-        const char *sysfs_path;
-        OrientationUp orientation;
-
-        sysfs_path = g_udev_device_get_sysfs_path (device);
-        g_debug ("Received uevent '%s' from '%s'", action, sysfs_path);
-
-        if (manager->priv->orientation_lock)
-                return;
-
-        if (g_str_equal (action, "change") == FALSE)
-                return;
-
-        if (g_strcmp0 (manager->priv->sysfs_path, sysfs_path) != 0)
-                return;
-
-        g_debug ("Received an event from the accelerometer");
-
-        orientation = get_orientation_from_device (device);
-        if (orientation != manager->priv->prev_orientation) {
-                manager->priv->prev_orientation = orientation;
-                g_debug ("Orientation changed to '%s', switching screen rotation",
-                         orientation_to_string (manager->priv->prev_orientation));
-
-                do_rotation (manager);
-        }
-}
-
 static void
 orientation_lock_changed_cb (GSettings             *settings,
                              gchar                 *key,
@@ -316,7 +251,7 @@ orientation_lock_changed_cb (GSettings             *settings,
 {
         gboolean new;
 
-        new = g_settings_get_boolean (settings, key);
+        new = g_settings_get_boolean (settings, ORIENTATION_LOCK_KEY);
         if (new == manager->priv->orientation_lock)
                 return;
 
@@ -332,11 +267,59 @@ orientation_lock_changed_cb (GSettings             *settings,
         }
 }
 
+static void
+properties_changed (GDBusProxy *proxy,
+                    GVariant   *changed_properties,
+                    GStrv       invalidated_properties,
+                    gpointer    user_data)
+{
+        CsdOrientationManager *manager = user_data;
+        CsdOrientationManagerPrivate *p = manager->priv;
+        GVariant *v;
+        GVariantDict dict;
+
+        if (manager->priv->xrandr_proxy == NULL)
+                return;
+
+        if (changed_properties)
+                g_variant_dict_init (&dict, changed_properties);
+
+        if (changed_properties == NULL ||
+            g_variant_dict_contains (&dict, "HasAccelerometer")) {
+                v = g_dbus_proxy_get_cached_property (p->iio_proxy, "HasAccelerometer");
+                if (v == NULL) {
+                        g_debug ("Couldn't fetch HasAccelerometer property");
+                        return;
+                }
+                p->has_accel = g_variant_get_boolean (v);
+                if (!p->has_accel)
+                        p->prev_orientation = ORIENTATION_UNDEFINED;
+                g_variant_unref (v);
+        }
+
+        if (changed_properties == NULL ||
+            g_variant_dict_contains (&dict, "AccelerometerOrientation")) {
+                if (p->has_accel) {
+                        OrientationUp orientation;
+
+                        orientation = get_orientation_from_device (manager);
+                        if (orientation != p->prev_orientation) {
+                                p->prev_orientation = orientation;
+                                g_debug ("Orientation changed to '%s', switching screen rotation",
+                                         orientation_to_string (p->prev_orientation));
+
+                                do_rotation (manager);
+                        }
+                }
+        }
+}
+
 static void
 xrandr_ready_cb (GObject               *source_object,
                  GAsyncResult          *res,
                  CsdOrientationManager *manager)
 {
+        CsdOrientationManagerPrivate *p = manager->priv;
         GError *error = NULL;
 
         manager->priv->xrandr_proxy = g_dbus_proxy_new_finish (res, &error);
@@ -344,82 +327,11 @@ xrandr_ready_cb (GObject               *source_object,
                 g_warning ("Failed to get proxy for XRandR operations: %s", error->message);
                 g_error_free (error);
         }
-}
 
-static void
-on_bus_gotten (GObject               *source_object,
-               GAsyncResult          *res,
-               CsdOrientationManager *manager)
-{
-        GDBusConnection *connection;
-        GError *error = NULL;
-
-        connection = g_bus_get_finish (res, &error);
-        if (connection == NULL) {
-                g_warning ("Could not get session bus: %s", error->message);
-                g_error_free (error);
+        if (p->iio_proxy == NULL)
                 return;
-        }
-        manager->priv->connection = connection;
 
-        g_dbus_connection_register_object (connection,
-                                           CSD_ORIENTATION_DBUS_PATH,
-                                           manager->priv->introspection_data->interfaces[0],
-                                           NULL,
-                                           NULL,
-                                           NULL,
-                                           NULL);
-
-        g_dbus_proxy_new (manager->priv->connection,
-                          G_DBUS_PROXY_FLAGS_NONE,
-                          NULL,
-                          "org.cinnamon.SettingsDaemon",
-                          "/org/cinnamon/SettingsDaemon/XRANDR",
-                          "org.cinnamon.SettingsDaemon.XRANDR_2",
-                          NULL,
-                          (GAsyncReadyCallback) xrandr_ready_cb,
-                          manager);
-}
-
-static GUdevDevice *
-get_accelerometer (GUdevClient *client)
-{
-        GList *list, *listiio, *l;
-        GUdevDevice *ret, *parent;
-
-        /* Look for a device with the ID_INPUT_ACCELEROMETER=1 property */
-        ret = NULL;
-        list = g_udev_client_query_by_subsystem (client, "input");
-        listiio = g_udev_client_query_by_subsystem (client, "iio");
-        list = g_list_concat(list, listiio);
-        for (l = list; l != NULL; l = l->next) {
-                GUdevDevice *dev;
-
-                dev = l->data;
-                if (g_udev_device_get_property_as_boolean (dev, "ID_INPUT_ACCELEROMETER")) {
-                        ret = dev;
-                        continue;
-                }
-                g_object_unref (dev);
-        }
-        g_list_free (list);
-
-        if (ret == NULL)
-                return NULL;
-
-        /* Now walk up to the parent */
-        parent = g_udev_device_get_parent (ret);
-        if (parent == NULL)
-                return ret;
-
-        if (g_udev_device_get_property_as_boolean (parent, "ID_INPUT_ACCELEROMETER")) {
-                g_object_unref (ret);
-                ret = parent;
-        } else {
-                g_object_unref (parent);
-        }
-
-        return ret;
+        properties_changed (manager->priv->iio_proxy, NULL, NULL, manager);
 }
 
 static int read_sysfs_attr_as_int(const char *filename) {
@@ -469,54 +381,53 @@ static gboolean mpu_timer(CsdOrientationManager *manager) {
         return !manager->priv->orientation_lock;
 }
 
-static gboolean
-csd_orientation_manager_idle_cb (CsdOrientationManager *manager)
+static void
+iio_sensor_appeared_cb (GDBusConnection *connection,
+                        const gchar     *name,
+                        const gchar     *name_owner,
+                        gpointer         user_data)
 {
-        const char * const subsystems[] = { "input", NULL };
-        GUdevDevice *dev;
+        CsdOrientationManager *manager = user_data;
+        CsdOrientationManagerPrivate *p = manager->priv;
+        GError *error = NULL;
 
-        cinnamon_settings_profile_start (NULL);
+        p->iio_proxy = g_dbus_proxy_new_sync (connection,
+                                              G_DBUS_PROXY_FLAGS_NONE,
+                                              NULL,
+                                              "net.hadess.SensorProxy",
+                                              "/net/hadess/SensorProxy",
+                                              "net.hadess.SensorProxy",
+                                              NULL,
+                                              &error);
 
-        manager->priv->settings = g_settings_new (CONF_SCHEMA);
-        manager->priv->orientation_lock = g_settings_get_boolean (manager->priv->settings, ORIENTATION_LOCK_KEY);
-        g_signal_connect (G_OBJECT (manager->priv->settings), "changed::orientation-lock",
-                          G_CALLBACK (orientation_lock_changed_cb), manager);
-
-        manager->priv->client = g_udev_client_new (subsystems);
-        dev = get_accelerometer (manager->priv->client);
-        if (dev == NULL) {
-                g_debug ("Did not find an accelerometer");
-                cinnamon_settings_profile_end (NULL);
-                return FALSE;
-        }
-        manager->priv->sysfs_path = g_strdup (g_udev_device_get_sysfs_path (dev));
-        g_debug ("Found accelerometer at sysfs path '%s'", manager->priv->sysfs_path);
-
-        manager->priv->prev_orientation = get_orientation_from_device (dev);
-
-        /* Poll the sysfs attributes exposed by MPU6050 as it is not an uevent based input driver */
-        if (g_strcmp0 (g_udev_device_get_sysfs_attr (dev, "name"), "mpu6050") == 0) {
-                manager->priv->prev_orientation = ORIENTATION_NORMAL;
-                g_timeout_add_seconds(MPU_POLL_INTERVAL, (GSourceFunc) mpu_timer, manager);
-                mpu6050_accel_x = g_build_filename(manager->priv->sysfs_path, "in_accel_x_raw", NULL);
-                mpu6050_accel_y = g_build_filename(manager->priv->sysfs_path, "in_accel_y_raw", NULL);
-                is_mpu6050 = TRUE;
+        if (p->iio_proxy == NULL) {
+                g_warning ("Failed to access net.hadess.SensorProxy after it appeared");
+                return;
         }
 
-        g_object_unref (dev);
+        g_dbus_proxy_call_sync (p->iio_proxy,
+                                "ClaimAccelerometer",
+                                NULL,
+                                G_DBUS_CALL_FLAGS_NONE,
+                                -1,
+                                NULL, NULL);
 
-        /* Start process of owning a D-Bus name */
-        g_bus_get (G_BUS_TYPE_SESSION,
-                   NULL,
-                   (GAsyncReadyCallback) on_bus_gotten,
-                   manager);
+        g_signal_connect (G_OBJECT (manager->priv->iio_proxy), "g-properties-changed",
+                          G_CALLBACK (properties_changed), manager);
 
-        g_signal_connect (G_OBJECT (manager->priv->client), "uevent",
-                          G_CALLBACK (client_uevent_cb), manager);
+        properties_changed (manager->priv->iio_proxy, NULL, NULL, manager);
+}
 
-        cinnamon_settings_profile_end (NULL);
+static void
+iio_sensor_vanished_cb (GDBusConnection *connection,
+                        const gchar     *name,
+                        gpointer         user_data)
+{
+        CsdOrientationManager *manager = user_data;
 
-        return FALSE;
+        g_clear_object (&manager->priv->iio_proxy);
+        manager->priv->has_accel = FALSE;
+        manager->priv->prev_orientation = ORIENTATION_UNDEFINED;
 }
 
 gboolean
@@ -525,10 +436,28 @@ csd_orientation_manager_start (CsdOrientationManager *manager,
 {
         cinnamon_settings_profile_start (NULL);
 
-        manager->priv->start_idle_id = g_idle_add ((GSourceFunc) csd_orientation_manager_idle_cb, manager);
+        manager->priv->settings = g_settings_new (CONF_SCHEMA);
+        g_signal_connect (G_OBJECT (manager->priv->settings), "changed::" ORIENTATION_LOCK_KEY,
+                          G_CALLBACK (orientation_lock_changed_cb), manager);
+        manager->priv->orientation_lock = g_settings_get_boolean (manager->priv->settings, ORIENTATION_LOCK_KEY);
 
-        manager->priv->introspection_data = g_dbus_node_info_new_for_xml (introspection_xml, NULL);
-        g_assert (manager->priv->introspection_data != NULL);
+        g_dbus_proxy_new_for_bus (G_BUS_TYPE_SESSION,
+                                  G_DBUS_PROXY_FLAGS_NONE,
+                                  NULL,
+                                  CSD_DBUS_NAME,
+                                  CSD_DBUS_PATH "/XRANDR",
+                                  CSD_DBUS_BASE_INTERFACE ".XRANDR_2",
+                                  NULL,
+                                  (GAsyncReadyCallback) xrandr_ready_cb,
+                                  manager);
+
+        manager->priv->watch_id = g_bus_watch_name (G_BUS_TYPE_SYSTEM,
+                                                    "net.hadess.SensorProxy",
+                                                    G_BUS_NAME_WATCHER_FLAGS_NONE,
+                                                    iio_sensor_appeared_cb,
+                                                    iio_sensor_vanished_cb,
+                                                    manager,
+                                                    NULL);
 
         cinnamon_settings_profile_end (NULL);
 
@@ -542,24 +471,28 @@ csd_orientation_manager_stop (CsdOrientationManager *manager)
 
         g_debug ("Stopping orientation manager");
 
-        if (p->settings) {
-                g_object_unref (p->settings);
-                p->settings = NULL;
+        if (p->watch_id > 0) {
+                g_bus_unwatch_name (p->watch_id);
+                p->watch_id = 0;
         }
 
-        if (p->sysfs_path) {
-                g_free (p->sysfs_path);
-                p->sysfs_path = NULL;
+        if (p->iio_proxy) {
+                g_dbus_proxy_call_sync (p->iio_proxy,
+                                        "ReleaseAccelerometer",
+                                        NULL,
+                                        G_DBUS_CALL_FLAGS_NONE,
+                                        -1,
+                                        NULL, NULL);
+                g_clear_object (&p->iio_proxy);
         }
 
-        if (p->introspection_data) {
-                g_dbus_node_info_unref (p->introspection_data);
-                p->introspection_data = NULL;
-        }
+        g_clear_object (&p->xrandr_proxy);
+        g_clear_object (&p->settings);
+        p->has_accel = FALSE;
 
-        if (p->client) {
-                g_object_unref (p->client);
-                p->client = NULL;
+        if (p->cancellable) {
+                g_cancellable_cancel (p->cancellable);
+                g_clear_object (&p->cancellable);
         }
 }
 
@@ -575,10 +508,7 @@ csd_orientation_manager_finalize (GObject *object)
 
         g_return_if_fail (orientation_manager->priv != NULL);
 
-        if (orientation_manager->priv->start_idle_id != 0) {
-                g_source_remove (orientation_manager->priv->start_idle_id);
-                orientation_manager->priv->start_idle_id = 0;
-        }
+        csd_orientation_manager_stop (orientation_manager);
 
         G_OBJECT_CLASS (csd_orientation_manager_parent_class)->finalize (object);
 }
